/* Routines for performing Ainsworth's text-to-sound.
   Author J. Zobel, jz@cs.rmit.edu.au, September 1995.
*/

#include <stdio.h>

#define BUFLEN	1024

void	loadrules();
int	ainsworth(char *, char *);

main(int argc, char *argv[])
{
    char	in[BUFLEN], out[BUFLEN];
    char	*gets();
    int		ret;

    loadrules();

    if( argc == 1 )
    {
	/* printf("> "); */
	while( gets(in) != NULL )
	{
	    ret = ainsworth(in, out);
	    printf("%s %s\n", in, out);
	    /* printf("> "); */
	}
	/* printf("\n"); */
    }
    else
    {
	ret = ainsworth(argv[1], out);
	printf("%s %s\n", argv[1], out);
    }

    exit(0);

}


#define RULEFILE	"/usr/local/ipa/ipa/ainsworth.dfn"
#define BUFLEN		1024
#define	LIM		10
#define true		1
#define false		0


    /* note that 'y' is both consonant and vowel */
#define ISCONS(c)	(  c == 'b' || c == 'c' || c == 'd' || c == 'f' \
			|| c == 'g' || c == 'h' || c == 'j' || c == 'k' \
			|| c == 'l' || c == 'm' || c == 'n' || c == 'p' \
			|| c == 'q' || c == 'r' || c == 's' || c == 't' \
			|| c == 'v' || c == 'w' || c == 'x' || c == 'y' \
			|| c == 'z' )

#define ISVOWEL(c)	(  c == 'a' || c == 'e' || c == 'i' || c == 'o' \
			|| c == 'u' || c == 'y' )


typedef struct rulerec
{
    char	pref[LIM], body[LIM], suf[LIM], phones[LIM];
    int		plen, blen, slen, Plen;
    struct rulerec *next;
} RULE;

/* Note that fixed-length arrays are smaller than variable-length
   malloc'ed ones */


RULE	*rules[256];


/* Load rules from named file.  Each rule is of the form

	pattern phonemes

   where pattern is of the form

	<pref>(<body>)<suf>

   <pref> and <suf> are patterns that must bound the <body>, which is
   the part that corresponds to the phonemes.  <pref> may start with
   ^ and <suf> may end with $.  Lower-case letters must be matched
   exactly.  There are also wildcards:

	C	match any number n of consonants, n>0
	V	match any number n of vowels, n>0

    The phoneme string is uninterpreted, except that `-' represents
    a empty string
*/
void
loadrules()
{
    FILE	*fp, *fopen();
    char	*p, inbuf[BUFLEN], *fgets();
    char	c, *malloc();
    RULE	*tmp;
    int		i;
    RULE	*tails[256];

    if( ( fp = fopen(RULEFILE, "r") ) == NULL )
    {
	fprintf(stderr, "Can't open %s for reading.\n", RULEFILE);
	exit(1);
    }

    for( i=0 ; i<256 ; i++ )
	tails[i] = rules[i] = NULL;

    while( ( p = fgets(inbuf, BUFLEN, fp) ) != NULL )
    {
	tmp = (RULE *) malloc( sizeof(RULE) );

	    /* extract the prefix */
	for( i=0 ; *p != '\0' && *p != '(' ; i++, p++ )
	    tmp->pref[i] = *p;
	tmp->pref[i] = '\0';
	tmp->plen = i;
	if( *p != '(' )
	{
	    fprintf(stderr, "Badly formatted rule, error 1: %s", inbuf);
	    exit(1);
	}

	    /* extract the middle */
	p++;
	for( i=0 ; *p != '\0' && *p != ')' ; i++, p++ )
	    tmp->body[i] = *p;
	tmp->body[i] = '\0';
	tmp->blen = i;
	if( *p != ')' )
	{
	    fprintf(stderr, "Badly formatted rule, error 2: %s", inbuf);
	    exit(1);
	}

	p++;
	for( i=0 ; *p != '\0' && *p != ' ' && *p != '\t' ; i++, p++ )
	    tmp->suf[i] = *p;
	tmp->suf[i] = '\0';
	tmp->slen = i;
	if( *p != ' ' && *p != '\t' )
	{
	    fprintf(stderr, "Badly formatted rule, error 3: %s", inbuf);
	    exit(1);
	}

	p++;
	i = 0;
	if( p[i] != '-' ) 
	    for( i=0 ; *p != '\0' && *p != '\n' ; i++, p++ )
		tmp->phones[i] = *p;
	tmp->phones[i] = '\0';
	tmp->Plen = i;

	tmp->next = NULL;

	c = *(tmp->body);	/* index rules on first char to be replaced */
	    /* add rules at end of list, to maintain priority order */
	if( tails[c] == NULL )
	    tails[c] = rules[c] = tmp;
	else
	{
	    tails[c]->next = tmp;
	    tails[c] = tmp;
	}
    }

/* #define DEBUG */
#ifdef DEBUG
for( i=0 ; i<256 ; i++ )
    for( tmp=rules[i] ; tmp!=NULL ; tmp=tmp->next )
    {
	printf("|%s|%s|%s| .... |%s|\n",
		tmp->pref, tmp->body, tmp->suf, tmp->phones);
    }
#endif /* DEBUG */

    fclose(fp);
    return;
}


/* Process a string with regard to the rules.  Proceed left to right,
   applying the rules for the current character in priority order.
   Return true if ok, false on error.
*/
int
ainsworth(char *in, char *out)
{
    int		pos, i, j, outlen;
    int		len = strlen(in);
    int		match;
    RULE	*rule;

    for( outlen=pos=0 ; pos<len ; )
    {
	match = false;
	for( rule=rules[in[pos]] ; !match && rule!=NULL ; rule=rule->next )
	{
	    match = true;

		/* match up the body */
	    for( j=pos, i=0 ; match && rule->body[i]==in[j] ; i++, j++ )
		;
	    if( rule->body[i] != '\0' )
		match = false;

		/* match up the prefix */
	    for( j=pos-1, i=rule->plen-1 ; match && j>=-1 && i>=0 ; i-- )
	    {
		if( rule->pref[i]=='^' )
		{
		    if( j!=-1 )
			match = false;
		}
		else if( j>=0 && rule->pref[i]=='C' && ISCONS(in[j]) )
		    for( j-- ; j>=0 && ISCONS(in[j]) ; j-- )
			;
		else if( j>=0 && rule->pref[i]=='V' && ISVOWEL(in[j]) )
		    for( j-- ; j>=0 && ISVOWEL(in[j]) ; j-- )
			;
		else if( j>=0 && rule->pref[i]==in[j] )
		    j--;
		else
		    match = false;
	    }
	    if( i>=0 )
		match = false;

		/* match up the suffix */
	    for( j=pos+rule->blen, i=0 ; match && j<=len && i<rule->slen ; i++ )
	    {
		if( rule->suf[i]=='$' )
		{
		    if( j!=len )
			match = false;
		}
		else if( j<len && rule->suf[i]=='C' && ISCONS(in[j]) )
		    for( j++ ; j<len && ISCONS(in[j]) ; j++ )
			;
		else if( j<len && rule->suf[i]=='V' && ISVOWEL(in[j]) )
		    for( j++ ; j<len && ISVOWEL(in[j]) ; j++ )
			;
		else if( j<len && rule->suf[i]==in[j] )
		    j++;
		else
		    match = false;
	    }
	    if( i<rule->slen )
		match = false;

#ifdef DEBUG
printf("trying |%s|%s|%s| .... |%s|",
	rule->pref, rule->body, rule->suf, rule->phones);
if( match) printf(" -- ok");
printf("\n");
#endif /* DEBUG */

	    if( match )
	    {
		for( i=0 ; i<rule->Plen ; i++, outlen++ )
		    out[outlen] = rule->phones[i];
		pos += rule->blen;
	    }
	}
	if( !match )
	    return(false);
    }

    out[outlen] = '\0';
    return(true);
}
